/*  shmcopy --  main */
#include "share_ex.h"

#define IFLAGS (IPC_CREAT | IPC_EXCL)
#define ERR ((struct databuf *) -1)

static int shmid1, shmid2, semid;

/*   p()  v()    */
struct sembuf p1 = {0, -1, 0}, p2 = {1, -1, 0};
struct sembuf v1 = {0, 1, 0}, v2 = {1, 1, 0};

int 
fatal (char *s)
{
  perror (s);
  exit (1);
}

/*  writer --   */
void 
writer (int semid, struct databuf *buf1,
	struct databuf *buf2)
{
  for (;;)
    {
      semop (semid, &p1, 1);
      semop (semid, &v2, 1);

      if (buf1->d_nread <= 0)
	return;

      write (1, buf1->d_buf, buf1->d_nread);
      semop (semid, &p2, 1);
      semop (semid, &v1, 1);

      if (buf2->d_nread <= 0)
	return;

      write (1, buf2->d_buf, buf2->d_nread);
    }

}

/*  reader --     */
void 
reader (int semid, struct databuf *buf1,
	struct databuf *buf2)
{
  for (;;)
    {
      /*    buf1 */
      buf1->d_nread = read (0, buf1->d_buf, SIZ);

      /*   */
      semop (semid, &v1, 1);
      semop (semid, &p2, 1);

      /*   writer   . */
      if (buf1->d_nread <= 0)
	return;

      buf2->d_nread = read (0, buf2->d_buf, SIZ);

      semop (semid, &v2, 1);
      semop (semid, &p1, 1);

      if (buf2->d_nread <= 0)
	return;
    }
}

void 
getseg (struct databuf **p1, struct databuf **p2)
{

  /*     */
  if ((shmid1 = shmget (SHMKEY1, sizeof (struct databuf), 0600 | IFLAGS)) == -1)
      fatal ("shmget");

  if ((shmid2 = shmget (SHMKEY2, sizeof (struct databuf),
			0600 | IFLAGS)) == -1)
      fatal ("shmget");

  /*    . */
  if ((*p1 = (struct databuf *) shmat (shmid1, 0, 0)) == ERR)
    fatal ("shmat");

  if ((*p2 = (struct databuf *) shmat (shmid2, 0, 0)) == ERR)
    fatal ("shmat");

}

int 
getsem (void)			/*    */
{
  union semun x;

  x.val = 0;

  /*     */
  if ((semid = semget (SEMKEY, 2, 0600 | IFLAGS)) == -1)
    fatal ("semget");

  /*    */

  if (semctl (semid, 0, SETVAL, x) == -1)
    fatal ("semctl");

  if (semctl (semid, 1, SETVAL, x) == -1)
    fatal ("semctl");

  return (semid);
}

 /*    
  *    
  */
void 
remobj (void)
{

  if (shmctl (shmid1, IPC_RMID, NULL) == -1)
    fatal ("shmctl");

  if (shmctl (shmid2, IPC_RMID, NULL) == -1)
    fatal ("shmctl");

  if (semctl (semid, 0, IPC_RMID, NULL) == -1)
    fatal ("semctl");
}

main ()
{
  int semid;
  pid_t pid;
  struct databuf *buf1, *buf2;

  /*   . */
  semid = getsem ();

  /*      . */
  getseg (&buf1, &buf2);

  switch (pid = fork ())
    {
    case -1:
      fatal ("fork");
    case 0:			/*   */
      writer (semid, buf1, buf2);
      remobj ();
      break;
    default:			/*   */
      reader (semid, buf1, buf2);
      break;
    }

  exit (0);
}
